#define std_io_InputTube_IMPORT
#define std_io_OutputTube_IMPORT
#define std_lang_ByteArray_IMPORT
#define std_lang_Chunk_IMPORT
#define std_lang_Object_IMPORT
#define std_util_LinkedList_IMPORT
#define std_util_Pool_IMPORT

#include <string.h>
#include <stdlib.h>
#include <std/io/impl/InputTube.h>
#include <std/io/impl/OutputTube.h>
#include <std/lang/ByteArray.h>
#include <std/lang/Chunk.h>
#include <std/lang/Object.h>
#include <std/util/LinkedList.h>
#include <std/util/impl/Pool.h>

union Pool
Pool_create(void)
{
    struct impl_Pool *impl;

    if ((impl = malloc(sizeof(struct impl_Pool))) == NULL) {
	return (union Pool)Object_NULL;
    }
    if ((impl->bucket = LinkedList_create()).super.object == NULL) {
	free(impl);
	return (union Pool)Object_NULL;
    }
    impl->size = 0;
    return (union Pool)impl;
}

size_t
Pool_getSize(union Pool p)
{
    struct impl_Pool *impl = p.impl;

    return impl->size;
}

static void
copyPool(unsigned char *d, struct LinkedList_Node *node)
{
    struct ByteArray *e;
    struct Chunk src;

    for (; node != NULL; node = node->next) {
	e = node->value;
	src = ByteArray_getChunk(e);
	memcpy(d, src.data, src.size);
	d += src.size;
    }
}

struct ByteArray *
Pool_createByteArray(union Pool p)
{
    struct impl_Pool *impl = p.impl;
    struct ByteArray *array;

    if ((array = ByteArray_create(impl->size)) == NULL)
	return NULL;
    copyPool(ByteArray_getData(array), LinkedList_getHeadNode(impl->bucket));
    return array;
}

void
Pool_free(union Pool p)
{
    struct impl_Pool *impl = p.impl;
    struct ByteArray *a;

    while ((a = LinkedList_getFirst(impl->bucket)) != NULL)
	ByteArray_free(a);
    LinkedList_free(impl->bucket);
    free(impl);
}

struct impl_PoolOutputTube {
    union {
        struct impl_OutputTube outputTube;
    } super;
    struct impl_Pool *pool;
};

union POT {
    struct impl_OutputTube *outputTube;
    struct impl_PoolOutputTube *impl;
} __attribute__((__transparent_union__));

static ssize_t
POT_write(union POT this, const void *data, size_t size)
{
    struct impl_PoolOutputTube *impl = this.impl;
    struct impl_Pool *pool = impl->pool;
    struct ByteArray *a;

    if ((a = ByteArray_createWithData(data, size)) == NULL) {
	return -1;
    }
    if (LinkedList_append(pool->bucket, a) < 0) {
	ByteArray_free(a);
	return -1;
    }
    pool->size += size;
    return size;
}

static int
POT_close(union POT this)
{
    struct impl_PoolOutputTube *impl = this.impl;

    free(impl);
    return 0;
}

static const struct OutputTube_Methods POT_METHODS = {
    .write = POT_write,
    .close = POT_close};

union OutputTube
Pool_createOutputTube(union Pool p)
{
    struct impl_PoolOutputTube *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
        return (union OutputTube)Object_NULL;
    }
    OutputTube_initialize(&impl->super.outputTube, &POT_METHODS);
    impl->pool = p.impl;
    return (union OutputTube)&impl->super.outputTube;
}

struct impl_PoolInputTube {
    union {
        struct impl_InputTube inputTube;
    } super;
    union LinkedList bucket;
    ssize_t (*read)(struct impl_PoolInputTube *, void *data, size_t);
    struct LinkedList_Node *node;
    size_t offset;
};

union PIT {
    struct impl_InputTube *inputTube;
    struct impl_PoolInputTube *impl;
} __attribute__((__transparent_union__));

static ssize_t
read2nd(struct impl_PoolInputTube *impl, void *data, size_t size)
{
    struct ByteArray *a;
    struct Chunk bucket;
    size_t len;

    if (impl->node == NULL)
	return 0;
    a = impl->node->value;
    bucket = ByteArray_getChunk(a);
    if ((len = bucket.size - impl->offset) > size) {
	len = size;
    }
    memcpy(data, (unsigned char *)bucket.data + impl->offset, len);
    impl->offset += len;
    if (impl->offset == bucket.size) {
	impl->node = impl->node->next;
	impl->offset = 0;
    }
    return len;
}

static ssize_t
read1st(struct impl_PoolInputTube *impl, void *data, size_t size)
{
    impl->node = LinkedList_getHeadNode(impl->bucket);
    impl->offset = 0;
    impl->read = read2nd;
    return read2nd(impl, data, size);
}

static ssize_t
PIT_read(union PIT this, void *data, size_t size)
{
    struct impl_PoolInputTube *impl = this.impl;

    return impl->read(impl, data, size);
}

static int
PIT_close(union PIT this)
{
    struct impl_PoolInputTube *impl = this.impl;

    free(impl);
    return 0;
}

static const struct InputTube_Methods PIT_METHODS = {
    .read = PIT_read,
    .close = PIT_close};

union InputTube
Pool_createInputTube(union Pool p)
{
    struct impl_PoolInputTube *impl;

    if ((impl = malloc(sizeof(*impl))) == NULL) {
        return (union InputTube)Object_NULL;
    }
    InputTube_initialize(&impl->super.inputTube, &PIT_METHODS);
    impl->bucket = p.impl->bucket;
    impl->read = read1st;
    impl->node = NULL;
    impl->offset = 0;
    return (union InputTube)&impl->super.inputTube;
}
